package broker

import (
	"fmt"
	"log"

	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

type LifeCycleRunner struct {
	boshClient BoshClient
	plans      Plans
}

func NewLifeCycleRunner(
	boshClient BoshClient,
	plans Plans,
) LifeCycleRunner {
	return LifeCycleRunner{
		boshClient,
		plans,
	}
}

func (l LifeCycleRunner) GetTask(deploymentName string, operationData OperationData, planID string, logger *log.Logger) (boshclient.BoshTask, error) {
	if operationData.BoshContextID == "" || !validPostDeployOpType(operationData.OperationType) {
		return l.boshClient.GetTask(operationData.BoshTaskID, logger)
	}

	boshTasks, err := l.boshClient.GetTasksByContext(deploymentName, operationData.BoshContextID, logger)
	if err != nil {
		return boshclient.BoshTask{}, err
	}

	switch len(boshTasks) {
	case 0:
		return boshclient.BoshTask{}, fmt.Errorf("no tasks found for context id: %s", operationData.BoshContextID)
	case 1:
		errand := l.postDeployErrand(planID)
		task := boshTasks[0]
		if errand != "" && task.Done() {
			return l.runErrand(deploymentName, errand, operationData.BoshContextID, logger)
		} else {
			return task, nil
		}
	case 2:
		return boshTasks[0], nil
	default:
		return boshclient.BoshTask{},
			fmt.Errorf("unexpected tasks found with context id: %s, tasks: %s", operationData.BoshContextID, boshTasks.ToLog())
	}
}

func validPostDeployOpType(op OperationType) bool {
	return op == OperationTypeCreate || op == OperationTypeUpdate
}

func (l LifeCycleRunner) postDeployErrand(planID string) string {
	plan := l.plans.FindByID(planID)

	if plan.LifecycleErrands != nil {
		return plan.LifecycleErrands.PostDeploy
	}
	return ""
}

func (l LifeCycleRunner) runErrand(deploymentName, errand, contextID string, log *log.Logger) (boshclient.BoshTask, error) {
	taskID, err := l.boshClient.RunErrand(deploymentName, errand, contextID, log)
	if err != nil {
		return boshclient.BoshTask{}, err
	}

	task, err := l.boshClient.GetTask(taskID, log)
	if err != nil {
		return boshclient.BoshTask{}, err
	}

	return task, nil
}
