package broker_test

import (
	"errors"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/ginkgo/extensions/table"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

var _ = Describe("Lifecycle runner", func() {
	const (
		deploymentName       = "some-deployment"
		contextID            = "some-uuid"
		planID               = "some-plan-id"
		errand               = "some-errand"
		anotherErrand        = "another-errand"
		anotherPlanID        = "another-plan-id"
		planIDWithoutErrands = "without-errands-plan-id"
	)

	plans := broker.Plans{
		broker.Plan{
			ID: planID,
			LifecycleErrands: &broker.LifecycleErrands{
				PostDeploy: errand,
			},
		},
		broker.Plan{
			ID: anotherPlanID,
			LifecycleErrands: &broker.LifecycleErrands{
				PostDeploy: anotherErrand,
			},
		},
		broker.Plan{
			ID: planIDWithoutErrands,
		},
	}

	taskProcessing := boshclient.BoshTask{ID: 1, State: boshclient.BoshTaskProcessing, Description: "snapshot deployment", Result: "result-1", ContextID: contextID}
	taskErrored := boshclient.BoshTask{ID: 2, State: boshclient.BoshTaskError, Description: "snapshot deployment", Result: "result-1", ContextID: contextID}
	taskComplete := boshclient.BoshTask{ID: 3, State: boshclient.BoshTaskDone, Description: "snapshot deployment", Result: "result-1", ContextID: contextID}

	var deployRunner broker.LifeCycleRunner
	var logger *log.Logger

	BeforeEach(func() {
		deployRunner = broker.NewLifeCycleRunner(
			boshClient,
			plans,
		)

		logger = loggerFactory.NewLogger()
	})

	Describe("post-deploy errand", func() {
		Context("when operation data has a context id", func() {
			operationData := broker.OperationData{BoshContextID: contextID, OperationType: broker.OperationTypeCreate}

			Context("when a first task is incomplete", func() {
				BeforeEach(func() {
					boshClient.GetTasksByContextReturns(boshclient.BoshTasks{taskProcessing}, nil)
				})

				It("returns the processing task", func() {
					task, _ := deployRunner.GetTask(deploymentName, operationData, planID, logger)
					Expect(task.State).To(Equal(boshclient.BoshTaskProcessing))
				})
			})

			Context("when a first task has errored", func() {
				BeforeEach(func() {
					boshClient.GetTasksByContextReturns(boshclient.BoshTasks{taskErrored}, nil)
				})

				It("returns the errored task", func() {
					task, _ := deployRunner.GetTask(deploymentName, operationData, planID, logger)
					Expect(task.State).To(Equal(boshclient.BoshTaskError))
				})
			})

			Context("when a first task cannot be found", func() {
				BeforeEach(func() {
					boshClient.GetTasksByContextReturns(boshclient.BoshTasks{}, nil)
				})

				It("returns an error", func() {
					_, err := deployRunner.GetTask(deploymentName, operationData, planID, logger)
					Expect(err).To(MatchError("no tasks found for context id: " + contextID))
				})
			})

			Context("when a first task is complete", func() {
				var task boshclient.BoshTask

				Context("and a post deploy is not configured", func() {
					BeforeEach(func() {
						deployRunner = broker.NewLifeCycleRunner(boshClient, plans)
						boshClient.GetTasksByContextReturns(boshclient.BoshTasks{taskComplete}, nil)
						task, _ = deployRunner.GetTask(deploymentName, operationData, planIDWithoutErrands, logger)
					})

					It("returns the completed task", func() {
						Expect(task.State).To(Equal(boshclient.BoshTaskDone))
					})

					It("does not run a post deploy errand", func() {
						Expect(boshClient.RunErrandCallCount()).To(Equal(0))
					})
				})

				Context("and a post deploy is configured", func() {
					BeforeEach(func() {
						boshClient.GetTasksByContextReturns(boshclient.BoshTasks{taskComplete}, nil)
						boshClient.RunErrandReturns(taskProcessing.ID, nil)
						boshClient.GetTaskReturns(taskProcessing, nil)
						task, _ = deployRunner.GetTask(deploymentName, operationData, planID, logger)
					})

					It("runs the post deploy errand", func() {
						Expect(boshClient.RunErrandCallCount()).To(Equal(1))
					})

					It("runs the correct errand", func() {
						_, errandName, _, _ := boshClient.RunErrandArgsForCall(0)
						Expect(errandName).To(Equal(errand))
					})

					It("runs the errand with the correct contextID", func() {
						_, _, ctxID, _ := boshClient.RunErrandArgsForCall(0)
						Expect(ctxID).To(Equal(contextID))
					})

					It("returns the post deploy errand processing task", func() {
						Expect(task.ID).To(Equal(taskProcessing.ID))
						Expect(task.State).To(Equal(boshclient.BoshTaskProcessing))
					})

					Context("and a post deploy errand is incomplete", func() {
						BeforeEach(func() {
							boshClient.GetTasksByContextReturns(boshclient.BoshTasks{taskProcessing, taskComplete}, nil)
							task, _ = deployRunner.GetTask(deploymentName, operationData, planID, logger)
						})

						It("returns the processing task", func() {
							Expect(task.State).To(Equal(boshclient.BoshTaskProcessing))
						})

						It("does not run a post deploy errand again", func() {
							Expect(boshClient.RunErrandCallCount()).To(Equal(1))
						})
					})

					Context("and a post deploy errand is complete", func() {
						BeforeEach(func() {
							boshClient.GetTasksByContextReturns(boshclient.BoshTasks{taskComplete, taskComplete}, nil)
							task, _ = deployRunner.GetTask(deploymentName, operationData, planID, logger)
						})

						It("returns the complete task", func() {
							Expect(task.State).To(Equal(boshclient.BoshTaskDone))
						})

						It("does not run a post deploy errand again", func() {
							Expect(boshClient.RunErrandCallCount()).To(Equal(1))
						})
					})

					Context("and the post deploy errand fails", func() {
						BeforeEach(func() {
							boshClient.GetTasksByContextReturns(boshclient.BoshTasks{taskErrored, taskComplete}, nil)
							task, _ = deployRunner.GetTask(deploymentName, operationData, planID, logger)
						})

						It("returns the failed task", func() {
							Expect(task.State).To(Equal(boshclient.BoshTaskError))
						})

						It("does not run a post deploy errand again", func() {
							Expect(boshClient.RunErrandCallCount()).To(Equal(1))
						})
					})

					Context("and when running the errand errors", func() {
						BeforeEach(func() {
							boshClient.RunErrandReturns(0, errors.New("some errand err"))
						})

						It("returns an error", func() {
							_, err := deployRunner.GetTask(deploymentName, operationData, planID, logger)
							Expect(err).To(MatchError("some errand err"))
						})
					})

					Context("and the errand task cannot be found", func() {
						BeforeEach(func() {
							boshClient.GetTaskReturns(boshclient.BoshTask{}, errors.New("some err"))
						})

						It("returns an error", func() {
							_, err := deployRunner.GetTask(deploymentName, operationData, planID, logger)
							Expect(err).To(MatchError("some err"))
						})
					})
				})
			})

			Context("when getting tasks errors", func() {
				BeforeEach(func() {
					boshClient.GetTasksByContextReturns(boshclient.BoshTasks{}, errors.New("some err"))
				})

				It("returns an error", func() {
					_, err := deployRunner.GetTask(deploymentName, operationData, planID, logger)
					Expect(err).To(MatchError("some err"))
				})
			})
		})

		Context("when operation data has no context id", func() {
			operationData := broker.OperationData{BoshTaskID: taskProcessing.ID, OperationType: broker.OperationTypeCreate}

			BeforeEach(func() {
				boshClient.GetTaskReturns(taskProcessing, nil)
			})

			It("calls get tasks with the correct id", func() {
				deployRunner.GetTask(deploymentName, operationData, planID, logger)

				Expect(boshClient.GetTaskCallCount()).To(Equal(1))
				actualTaskID, _ := boshClient.GetTaskArgsForCall(0)
				Expect(actualTaskID).To(Equal(taskProcessing.ID))
			})

			It("returns the processing task", func() {
				task, _ := deployRunner.GetTask(deploymentName, operationData, planID, logger)

				Expect(task).To(Equal(taskProcessing))
			})

			It("does not error", func() {
				_, err := deployRunner.GetTask(deploymentName, operationData, planID, logger)

				Expect(err).ToNot(HaveOccurred())
			})

			Context("and bosh client returns an error", func() {
				BeforeEach(func() {
					boshClient.GetTaskReturns(boshclient.BoshTask{}, errors.New("no task for you"))
				})

				It("returns the error", func() {
					_, err := deployRunner.GetTask(deploymentName, operationData, planID, logger)

					Expect(err).To(MatchError("no task for you"))
				})
			})
		})

		DescribeTable("operation type",
			func(operationType broker.OperationType, errandRuns bool) {
				operationData := broker.OperationData{OperationType: operationType, BoshContextID: contextID}
				boshClient.GetTasksByContextReturns(boshclient.BoshTasks{taskComplete}, nil)
				deployRunner.GetTask(deploymentName, operationData, planID, logger)

				if errandRuns {
					Expect(boshClient.RunErrandCallCount()).To(Equal(1))
				} else {
					Expect(boshClient.RunErrandCallCount()).To(Equal(0))
				}
			},
			Entry("create runs errand", broker.OperationTypeCreate, true),
			Entry("update runs errand", broker.OperationTypeUpdate, true),
			Entry("delete does not run errand", broker.OperationTypeDelete, false),
			Entry("upgrade does not run errand", broker.OperationTypeUpgrade, false),
		)
	})
})
