package broker

import (
	"context"
	"encoding/json"
	"fmt"
	"log"

	"github.com/pborman/uuid"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/brokercontext"
)

func (b *Broker) Update(
	ctx context.Context,
	instanceID string,
	details brokerapi.UpdateDetails,
	asyncAllowed bool,
) (brokerapi.UpdateServiceSpec, error) {
	b.deploymentLock.Lock()
	defer b.deploymentLock.Unlock()

	requestID := uuid.New()
	ctx = brokercontext.New(ctx, string(OperationTypeUpdate), requestID, b.serviceOffering.Name, instanceID)
	logger := b.loggerFactory.NewLoggerWithContext(ctx)

	if !asyncAllowed {
		return brokerapi.UpdateServiceSpec{}, brokerapi.ErrAsyncRequired
	}

	errs := func(err DisplayableError) (brokerapi.UpdateServiceSpec, error) {
		logger.Println(err)
		return brokerapi.UpdateServiceSpec{IsAsync: true}, err.ErrorForCFUser()
	}

	planConfig := b.serviceOffering.FindPlanById(details.PlanID)
	if planConfig == nil {
		return errs(NewDisplayableError(fmt.Errorf("plan %s not found", details.PlanID), fmt.Errorf("finding plan ID %s", details.PlanID)))
	}

	err := b.validatePlanQuota(planConfig, details, logger)
	switch err := err.(type) {
	case DisplayableError:
		return errs(err)
	case error:
		return errs(NewGenericError(ctx, err))
	}

	logger.Printf("updating instance %s", instanceID)
	detailsMap, err := convertToMap(details)
	if err != nil {
		return errs(NewGenericError(ctx, err))
	}

	var boshContextID string
	var operationDataPlanID string
	if planConfig.LifecycleErrands != nil {
		boshContextID = uuid.New()
		operationDataPlanID = details.PlanID
	}

	boshTaskID, err := b.redeploy(instanceID, details.PlanID, OperationTypeUpdate, detailsMap, &details.PreviousValues.PlanID, boshContextID, logger)

	switch err := err.(type) {
	case boshclient.RequestError:
		return errs(NewBoshRequestError("update", fmt.Errorf("error deploying instance: %s", err)))
	case DisplayableError:
		return errs(err)
	case error:
		return errs(
			NewGenericError(ctx, fmt.Errorf("error deploying instance: %s", err)),
		)
	}

	ctx = brokercontext.WithBoshTaskID(ctx, boshTaskID)

	operationData, err := json.Marshal(OperationData{
		BoshTaskID:    boshTaskID,
		OperationType: OperationTypeUpdate,
		BoshContextID: boshContextID,
		PlanID:        operationDataPlanID,
	})
	if err != nil {
		return errs(NewGenericError(ctx, err))
	}

	return brokerapi.UpdateServiceSpec{IsAsync: true, OperationData: string(operationData)}, nil
}

func (b *Broker) validatePlanQuota(planConfig *Plan, details brokerapi.UpdateDetails, logger *log.Logger) error {
	if details.PreviousValues.PlanID == details.PlanID || planConfig.Quotas == nil {
		return nil
	}

	count, err := b.cfClient.CountInstancesOfPlan(details.ServiceID, details.PlanID, logger)
	if err != nil {
		return fmt.Errorf("error counting instances of plan: %s", err)
	}

	if planConfig.Quotas != nil && count >= planConfig.Quotas.ServiceInstanceLimit {
		return NewDisplayableError(brokerapi.ErrPlanQuotaExceeded, fmt.Errorf("plan quota exceeded for plan ID %s", details.PlanID))
	}

	return nil
}
